#include <math.h>

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include "ball.h"
#include "main.h"

Ball::Ball(unsigned int id) : Littlecube(id) {
  attached = 0;
  //  printf("new ball (%d)\n",id);
}
Ball::~Ball() {}

void Ball::init(double x,double y) {
  struct ball_info bi;
  bi.id       = 0; // not used by process()
  bi.cls      = CLS_BALL;
  bi.ev       = EV_INIT;
  bi.x        = x;
  bi.y        = y;
  process((item_info*)&bi,0); // this calls LittleCube::init

  reset();

  //  if (!is_remote()) // the test optimises but is not strictly necessary
  poll();
}

void Ball::process(struct item_info *data,size_t len) {
  struct ball_info *bi = (struct ball_info*)data;

  switch (bi->ev) {
  case EV_INIT:
    Littlecube::init(bi->x,bi->y,0.2f,true,false);
    break;
  case EV_MOVED:
    move(bi->x,bi->y);
    break;
  }
}

void Ball::poll() {
  struct ball_info bi;
  bi.cls      = CLS_BALL;
  bi.ev       = EV_INIT;
  bi.x        = x;
  bi.y        = y;
  push((item_info*)&bi,sizeof(struct ball_info));
}

void Ball::reset() {
  r = 0.2f;
  set_delta_radius(0);
  ccount=0;
  tspeed = 4.2;
}

void Ball::attach(Pad *p) {
  set_speed(0,0);
  attached=p;
  p->capture(this);
}

void Ball::release(Pad *p) {
  if ( attached && (attached==p || !p) ) {
    int *dir = attached->get_direction();
    set_speed(dir[0]+dir[1], dir[0]+dir[1]);
    attached=0;
  }
}

bool Ball::update(double t,Level *lvl) {
  // printf("\nb:");
  Littlecube::update(t,lvl);
  // printf("\n");

  //    if (ccount>10) printf("%d\n",ccount);

  if (attached) {
    attached->capture(this);
    look_at_me();
    return true;
  }

  /* the ball constantly accelerates har har har */
  tspeed *= 1+t/100;

  /* Get closer from speed tspeed */

  double s = hypot(vx,vy); // current speed
  if (s>0.001) {
    double factor = (1-t) + (t)*(tspeed / s);
    vx *= factor+t/100; // change direction in case ball gets stuck
    vy *= factor;

    // printf("%f -> %f\n",s,tspeed);
  }

  if (has_vanished())
    return false;
  else {
    struct ball_info bi;
    bi.cls      = CLS_BALL;
    bi.ev       = EV_MOVED;
    bi.x        = x;
    bi.y        = y;
    push((item_info*)&bi,sizeof(struct ball_info));

    look_at_me();
    return true;
  }
}

void Ball::render() {

  glPushMatrix(); {
    glTranslatef(x-r,y-r,0);

    glColor3f(1,1-ccount/30.0,1-ccount/30.0);

    render_cube(2*r,2*r,2*r,true,r/1.5);

  /* Not actually a little cube, but rather a little octaedron :) */
    // glBegin(GL_TRIANGLE_FAN);
    // glNormal3f( 0,r, 0);
    // glVertex3f( 0,r, 0);

    // glNormal3f( r,0, 0);
    // glVertex3f( r,0, 0);

    // glNormal3f( 0,0, r);
    // glVertex3f( 0,0, r);

    // glNormal3f(-r,0, 0);
    // glVertex3f(-r,0, 0);

    // glNormal3f( 0,0,-r);
    // glVertex3f( 0,0,-r);

    // glNormal3f( 0,r, 0);
    // glVertex3f( 0,r, 0);
    // glEnd();

    // glBegin(GL_TRIANGLE_FAN);
    // glNormal3f( 0,-r, 0);
    // glVertex3f( 0,-r, 0);

    // glNormal3f( 0, 0,-r);
    // glVertex3f( 0, 0,-r);

    // glNormal3f(-r, 0, 0);
    // glVertex3f(-r, 0, 0);

    // glNormal3f( 0, 0, r);
    // glVertex3f( 0, 0, r);

    // glNormal3f( r, 0, 0);
    // glVertex3f( r, 0, 0);

    // glNormal3f( 0,r, 0);
    // glVertex3f( 0,r, 0);
    // glEnd();
  } glPopMatrix();
}
