#ifndef BALL_H
#define BALL_H

class Ball;

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include "littlecube.h"
#include "level.h"
#include "pad.h"

#define CLS_BALL 4

struct ball_info {
  unsigned int id;
  unsigned char cls; /* Must be CLS_BALL */
  unsigned char ev;
  /* Initial position */
  double x;
  double y;
};

class Ball : public Littlecube {
 public:
  Ball(unsigned int id);
  ~Ball();

  void init(double x,double y);

  void process(struct item_info *data,size_t len);
  void poll();

  void reset();

  /* Attach the ball to the given pad */
  void attach(Pad *p);

  /* release the ball from any pad it is attached to or, whenever a
     non-zero pad is given as parameter, only if the ball is attached
     to precisely that pad. */
  void release(Pad *p=0);

  /* Returns false if the ball got lost */
  bool update(double t,Level *lvl);

  void render();

 private:
  /* "target speed". The ball tries to run at that speed (slows down
     if it goes faster and accelerates if it's slower). That target
     speed gradually increases until it gets reset */
  double tspeed;

  /* Which pad, if any, the ball is attached to */
  Pad *attached;
};

#endif
