#include <math.h>

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include "main.h"
#include "base.h" 
#include "region.h"

/* CURrently visible area */
Region cur;
//double curxa,curxz,curya,curyz;

/* RECently requested area, must be included in CUR */
Region rec;
//double recxa,recxz,recya,recyz;

/* REQuested area to be shown. Enlarged using lookAt(), and must be
   included in CUR */
Region req;
//double reqxa, reqxz, reqya, reqyz;

/* TRAnsient requested area, that gets enlarged when lookAt is called
   with transient set to true, and reset every time */
Region tra;
//double traxa, traxz, traya, trayz;

/* Current camera position (computed using the above values) */
double camx, camy, camz;

/* Current angle, zero if we're looking along the z-axis */
double angx,angy;
double dangx,dangy; // derivatives...

/* extra movement backwards to make sure everything can be seen */
double extraz;

/* "camera reset time out". In how many seconds should REQ be moved
   to REC. */
double crto;

/* Instructs the camera to try having (x,y) visible. */
void lookAt(double x,double y, bool transient) {
  tra.include(x,y);
  if (!transient) {
    rec.include(x,y);
    req.include(x,y);
  }
}

void resetCam() {
  crto = 0;
  cur.reset();
  rec.reset();
  req.reset();
  tra.reset();

  angx=0; angy=0;
  dangx=0; dangy=0;
}

// void dontLook() {
//   crto = 0;
//   recxa = 100;
//   recxz = -100;
//   recya = 100;
//   recyz = -100;

//   reqxa = 100;
//   reqxz = -100;
//   reqya = 100;
//   reqyz = -100;
// }

void updateCam(double t) {
  crto -= t;
  if (crto<0) {
    crto = 5;
    rec.set_to(&req);
    req.reset();
  }

  /* Bring cur closer from the union of tra, rec and req. To have a
     nice hyberbolic tangent-oid, try to match the derivative with the
     remaining distance. */

  /* TARget */
  Region tar;
  tar.set_to(&tra);
  tar.include(&rec);
  tar.include(&req);

  cur.smooth_set_to(&tar,t);

  double *mid = cur.centre();
  camx = mid[0];
  camy = mid[1];
  camz = max(cur.width()*((double)screen_height/screen_width),
	     (cur.height()))+2;

  /* The camera position is determined by (cur..), and in case the
     target area is not visible when looking along the z axis, do some
     rotation to see the extreme area (where, presumably, the action
     is happening. In case stuff happens simultaneously in two
     opposite regions, just go for the middle) */
  // printf("y: ");
  dangx =  (1-t)*dangx + t* (tra.get_target_angle(false,camx,camz)-angx);
  dangy =  (1-t)*dangy + t* (tra.get_target_angle(true, camy,camz)-angy);
  angx += dangx*t;
  angy += dangy*t;
  // printf(": %f\n",angy);

  // printf("x: ");
  // printf(": %f\n",angx);

  /* Now, maybe rotation wasn't enough to see all the action, so move
     backwards until everything is visible */

  extraz *= 1-t;

  if (extraz<0) extraz = 0;
// #ifdef DEBUG
//   double debug = tra.get_extraz(true, camy,camz,angy);
//   extraz = max(extraz, max(debug,
// 			   tra.get_extraz(false,camx,camz,angx)));
//   printf("yextraz=%f\n",debug);
// #else
  extraz = max(extraz, max(tra.get_extraz(true, camy,camz,angy),
			   tra.get_extraz(false,camx,camz,angx)));
// #endif

  tra.reset();
}

void applyCam() {

  glTranslatef(0,0,-extraz);

  glRotatef(-angy RADIANS,1,0,0);
  glRotatef(angx RADIANS,0,1,0);

  /* The first two parameters center the currently viewable area and
     the third parameter moves forward or backwards to have said
     viewable area fit in the screen */
  glTranslatef(-camx,-camy,-camz);
}

void debugCam() {

  Region tar;
  tar.set_to(&tra);
  tar.include(&rec);
  tar.include(&req);

  // printf("cur: ");cur.print();
  // printf("tar: ");tar.print();

  glNormal3f(0,0,1);

  glColor3f(1,0,0); // target: red
  tar.paint();

  glColor3f(1,1,0); // current: yellow
  cur.paint();

  /* paint a grid */
  glBegin(GL_LINES);
  glColor3f(1,1,1);
  glNormal3f(0,0,1);
  for (int i=0;i<=10;i++) {
    glVertex3f(i,0,0);
    glVertex3f(i,10,0);
    glVertex3f(0,i,0);
    glVertex3f(10,i,0);
  }
  /* This point is supposed to be at the centre of the screen at all times */
  double mx = camx+tan(angx)*camz;
  double my = camy+tan(angy)*camz;

  glVertex3f(mx-0.2,my,0);
  glVertex3f(mx+0.2,my,0);
  glVertex3f(mx,my-0.2,0);
  glVertex3f(mx,my+0.2,0);

  glEnd();

  /* Complete field of view */
  double eax = camx + tan(angx-(PFOV*screen_ratio/2))*camz + 0.01; // left
  double ezx = camx + tan(angx+(PFOV*screen_ratio/2))*camz - 0.01; // right
  double eay = camy + tan(angy-(PFOV/2))*camz + 0.01; // bottom
  double ezy = camy + tan(angy+(PFOV/2))*camz - 0.01; // top

  glBegin(GL_LINE_LOOP);
  glVertex3f(eax,eay,0);
  glVertex3f(ezx,eay,0);
  glVertex3f(ezx,ezy,0);
  glVertex3f(eax,ezy,0);
  glEnd();

}
