#ifndef ITEM_H
#define ITEM_H

#include <string.h>

/* class Item; */

/* #include <stone.h> */

#define EV_INIT 1
#define EV_MOVED 2
#define EV_BROKE 3

/* Used when push()ing something over the network */
struct item_info {
  unsigned int id;
  /* CLS_something, as #defined in the corresponding .h file */	    
  unsigned char cls;
  unsigned char ev;
};

/** An object that you can see on the screen and that may be shared
    between several machines */

class Item {
 public:
  /* The constructor register_item()s this into the level.

     For remote objects, pass their identifier. For local objects,
     pass zero */
  Item(unsigned int id=0);
  ~Item();

  unsigned int get_id();

  /* Things that are remote should wait for remote events rather than
     updating themselves */
  bool is_remote();

  /* Broadcast a state update to whoever is listening.

     The structure you pass should begin like item_info, whose .id
     field gets filled by this method prior to transmitting.

     The resulting structure will be passed to the process() method of
     an item of the same class at the other end. */
  void push(item_info *data,size_t len);

  /* Request the object to push() its entire state (for new clients) */
  virtual void poll();

  /* Tell the item about a push() that was sent from its model on
     server-side. */
  virtual void process(struct item_info *data,size_t len);

 private:
  bool remote;

  unsigned int id;
};

Item *new_Item(struct item_info *data,size_t len);

#endif
