#include <stdlib.h>

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include "level.h"
#include "main.h"

#include "teleport.h"

Level::Level(bool passive) {
  this->passive = passive;
  next_item=1; // zero is reserved for "unassigned"
  levelnum=1;
  levelDone=false;
  fadeout=0;
  wid=0;
  hei=0;
  //  cells=0;
  //  stones=0;
  //  scount = 0;
  // pad = 0;
  // padcount = 0;
  favpad = 0;
  whichpads = 0;
  //  balls = 0;
  //  ballcap = 0;
  powerup = 0;
  lastcoll = 0;
}

Level::~Level() {
  clear();
}

void Level::clear() {
  for(list<Stone *>::iterator s=stones.begin();s != stones.end();++s) {
    delete (*s);
  }
  stones.clear();

  for (list<Ball *>::iterator b=balls.begin();b != balls.end();++b) {
    delete (*b);
  }
  balls.clear();

  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    delete (*p);
  }
  pads.clear();

 //  delete [] cells;
  //  if (stones) delete [] stones;
  // if (pad) delete [] pad;
  //  if (balls) delete [] balls;
  //  padcount = 0;
}

// Cell *Level::getCell(int x,int y) {
//   if (x<0 || x>=wid || y<0 || y>=hei) return 0;

//   return &(cells[y*wid+x]);
// }

/* Read a newline */
bool el(FILE *s) {
  char c;
  if (fread(&c,1,1,s)!=1) {
    printf("Unexpected end of file (newline expected)\n");
    return false;
  }
  if (c!='\n') {
    printf("Newline expected\n");
    return false;
  }
  return true;
}

bool Level::load(const char *lvl) {
  char fn[64];

  if (lvl) {
    if (snprintf(fn,64,"levels/%s",lvl)>=64) return false;
  } else {
    if (snprintf(fn,64,"levels/%d",levelnum)>=64) return false;
  }

  FILE *f = fopen(fn,"r");
  bool success = false;
  char c;
  Teleport *tele;
  Teleport *firstteleport = 0;
  Teleport *lastteleport = 0;
  int p; // avoid compiler warnings

  struct field_info fi;
  fi.id=0;
  fi.cls=CLS_FIELD;
  fi.pads = BOTTOMPAD;

  if (!f) {
    printf("couldn't open %s\n",fn);
    return false;
  }
  /* first line of a level file contains the field dimensions, and
     optionally extra pad specifications. */
  if (fscanf(f,"%d %d",&(fi.width),&(fi.height)) < 2) {
    printf("first line has illegal format\n");
    goto out;
  }

  /* Scan for additional flags on the first line */
  do {
    if (fread(&c,1,1,f)!=1) {
      printf("Unexpected eof\n");
      goto out;
    }
    switch (c) {
    case 'T': /* Toggle pad at the top of the screen */
      fi.pads ^= TOPPAD;
      break;
    case 'B': /* Toggle pad at the bottom of the screen */
      fi.pads ^= BOTTOMPAD;
      break;
    case 'L': /* Toggle pad at the left of the screen */
      fi.pads ^= LEFTPAD;
      break;
    case 'R': /* Toggle pad at the right of the screen */
      fi.pads ^= RIGHTPAD;
      break;
    case ' ': case '\n': /* Whitespace is ignored */
      break;
    default:
      printf("Unknown flag %.1s\n",&c);
      goto out;
    }
  } while (c != '\n');

  clear();
  process((item_info*) &fi,0);

  // padcount = ((fi->pads & TOPPAD)?1:0)+
  //   ((fi->pads & LEFTPAD)?1:0)+
  //   ((fi->pads & RIGHTPAD)?1:0)+
  //   ((fi->pads & BOTTOMPAD)?1:0);

  // pad = new Pad[padcount];

  //  int i=0; // index in the pad array
  p=1; // somethingPAD
  for (;p<=LASTPAD;p*=2) {
    if (whichpads & p) {
      Pad *pad = new Pad(0);
      pads.push_back(pad);
      pad->init(this,
		(p==LEFTPAD)?0:(p==RIGHTPAD)?wid:wid/2.0,
		p==BOTTOMPAD?0:(p==TOPPAD)?hei:hei/2.0,p);
      //      i++;
    }
  }
  favpad = *(pads.begin());

  //  printf("%d x %d\n",wid,hei);
  //  cells = new Cell[wid*hei];

  /* We have at most that many stones... */
  //  stones = new Stone*[wid*hei];

  to_break = 0;

  /* read cells */
  for (int y=hei-1;y>=0;y--) {
    for (int x=0;x<wid;x++) {
      Stone *s;
      if (fread(&c,1,1,f)!=1) {
	printf("Unexpected end of file\n");
	goto out;
      }

      s = new_stone(c);
      if (s) {
	s->init(c,x,y);
	if (s->breakable()) to_break++;
	if (c=='T') {
	  tele = (Teleport*)s;
	  if (lastteleport) {
	    tele->set_target(lastteleport);
	  } else {
	    firstteleport = tele;
	  }
	  lastteleport = tele;
	}
	add_stone(s);
      }
    }
    if (!el(f)) goto out;
    //    printf("\n");
  }

  if (firstteleport) {
    firstteleport->set_target(lastteleport);
  }

  //  ballcap = 1;
  //  balls = new Ball*[ballcap];
  extra_ball(favpad);

  success = true;

  /* LookAt the corners, to try having the full level visible */
  // lookAt(0,0);
  // lookAt(wid,hei);

 out:
  fclose(f);
  return success;
}

unsigned int Level::register_item(Item *i) {
  items.push_back(i);
  return i->get_id()? i->get_id() : next_item++;
}

void Level::process(struct item_info *data,size_t len) {
  struct field_info *fi = (struct field_info*)data;

  this->wid = fi->width;
  this->hei = fi->height;
  this->whichpads = fi->pads;
}

void Level::add_stone(Stone *s) {
  stones.push_back(s);
  //  stones[scount++] = s;
}

void Level::add_pad(Pad *p) {
  pads.push_back(p);
}

void Level::add_ball(Ball *b) {
  balls.push_back(b);
}

void Level::unregister(Item *i) {
  items.remove(i);
}

Item *Level::get_item(unsigned int id) {
  for(list<Item *>::iterator it=items.begin();it != items.end();++it) {
    if ((*it)->get_id()==id) return *it;
  }

  return NULL;
}

void Level::poll() {
  if (connection) {
    struct field_info fi;
    fi.id=0;
    fi.cls=CLS_FIELD;
    fi.ev=EV_INIT;
    fi.width = this->wid;
    fi.height = this->hei;
    fi.pads = whichpads;
    connection->push((item_info *) &fi,sizeof(struct field_info));

    for(list<Item *>::iterator it=items.begin();it != items.end();++it) {
      (*it)->poll();
    }
  }
}

void Level::skip() {
  levelDone = true;
}

  /** returns if the given trajectory intersects *any* object in the
      level (if so, return a double between 0 and 1 indicating where,
      sets n to the normal and v to the speed of the object on the
      path.

      Parameters:
      (x;y): Position of the object at time (now-t)

      (vx;vy): Speed of the object

      t: Time interval in which to search for collisions

      n: gets set to the normal vector of the obstacle in case of
         collision

      v: gets set to the speed of the obstacle in case of collision.

      trigger: Whether to call the trigger() method of stones in the
         path (if any)

      o: The object being on that trajectory

      Return value: Between 0 (t seconds ago) and 1 (now), when a
      collision occurred, or 2 if no collision occurred.
  **/
double Level::check_collision(double x,double y,double vx,double vy, double t,double *n, double *v, Pad **p, bool trigger, Littlecube *o) {
  double shortest=2; // shortest collision found so far
  double cur; // distance of current obstacle
  double curn[2];
  double curv[2];

  Stone *obstacle = 0; // which stone did we touch?

  /* first check the level boundaries. Not using the generic function
     because they are infinite walls */

  double ex = x+vx*t, ey = y+vy*t; // trajectory Endpoint

  if (!(whichpads & LEFTPAD)) {
    if (x>0 && ex<0) {
      cur = x/(x-ex);
      if (cur<shortest) { // (will always hold but keeping for
	// prettiness)
	shortest=cur;
	if (n) { n[0]=1; n[1]=0; }
	if (v) { v[0]=0; v[1]=0; }
	if (p) { *p=0; }
      }
    }
  }

  if (!(whichpads & RIGHTPAD)) {
    if (x<wid && ex>wid) {
      cur = (wid-x)/(ex-x);
      if (cur<shortest) {
	//      printf("wr");
	shortest=cur;
	if (n) { n[0]=-1; n[1]=0; }
	if (v) { v[0]=0; v[1]=0; }
	if (p) { *p=0; }
      }
    }
  }

  if (!(whichpads & TOPPAD)) {
    if (y<hei && ey>hei) {
      cur = (hei-y)/(ey-y);
      if (cur<shortest) {
	//      printf("wt");
	shortest=cur;
	if (n) { n[0]=0; n[1]=-1; }
	if (v) { v[0]=0; v[1]=0; }
	if (p) { *p=0; }
      }
    }
  }

  if (!(whichpads & BOTTOMPAD)) {
    if (y>0 && ey<0) {
      cur = y/(y-ey);
      if (cur<shortest) {
	//      printf("wt");
	shortest=cur;
	if (n) { n[0]=0; n[1]=1; }
	if (v) { v[0]=0; v[1]=0; }
	if (p) { *p=0; }
      }
    }
  }

  /* Now checking the stones */
  for(list<Stone *>::iterator s=stones.begin();s != stones.end();++s) {
    cur = (*s)->check_collision(x,y,vx*t,vy*t,curn,curv,trigger);
    if (cur<shortest) {
      //      printf("s");
      shortest=cur;
      obstacle = (*s);
      if (n) { n[0] = curn[0]; n[1] = curn[1]; }
      if (v) { v[0] = curv[0]; v[1] = curv[1]; }
      if (p) { *p=0; }
    }
  }

  /* Now checking the pads */

  for (list<Pad *>::iterator pad=pads.begin();pad != pads.end();++pad) {
    if ((*pad) == lastcoll) continue;
    cur = (*pad)->check_collision(x,y,vx,vy,t,curn,curv);
    if (cur<shortest) {
      if (trigger) (*pad)->trigger();
      //    printf("p");
      shortest=cur;
      obstacle = 0; // no stone got hit after all
      lastcoll = (*pad); // avoid double hitting of the pad by a ball
      if (trigger) favpad = *pad;
      if (n) { n[0] = curn[0]; n[1] = curn[1]; }
      if (v) { v[0] = curv[0]; v[1] = curv[1]; }
      if (p) { *p=(*pad); }
    }
  }

  /* WARN the lastcoll bit depends on pads being checked last */

  if (obstacle && trigger) {
    //    printf("%#x triggers %#x\n",o,s);
    if (obstacle->trigger(o)) {
      if ((--to_break)<=0) {
	levelDone=true;
      }
      /* broke a stone. Let's see if we get a powerup */
      if (favpad && (rand()%10 == 1) &&
	  !powerup) {
	powerup = new Powerup();
	int *pdir = favpad->get_direction();
	/* The powerup goes towards the pad so swap signs */
	pdir[0] = -pdir[0]; pdir[1] = -pdir[1];
	GLuint tx;
	switch (rand()%3) {
	case 0:
	  tx=wide_tex;
	  break;
	case 1:
	  tx=fire_tex;
	  break;
	case 2:
	  tx=extra_tex;
	}
	powerup->init(obstacle->getX(), obstacle->getY(),pdir, tx);
      }
    }
  }

  return shortest;
}

void Level::reset_coll() {
  lastcoll = 0;
}

bool Level::is_in_level(double x,double y) {
  return (x>-1 && x<wid+1 && y>-1 && y<hei+1);
  // if (x<-20 || x>wid+20 || y<-20 || y>hei+20) return 2;
  // else if (x<-1 || x>wid+1 || y<-1 || y>hei+1) return 1;
  // else return 0;
}

int Level::get_width() {
  return wid;
}

int Level::get_height() {
  return hei;
}

void Level::render() {

  /* A blue frame around the level */
  glBegin(GL_LINES);
  glNormal3f(0,0,1);
  glColor3f(0,-5,1);
  if (!(whichpads & LEFTPAD)) {
    glVertex3f(0,-5,0);
    glVertex3f(0,hei,0);
  }
  if (!(whichpads & TOPPAD)) {
    glVertex3f(0,hei,0);
    glVertex3f(wid,hei,0);
  }
  if (!(whichpads & RIGHTPAD)) {
    glVertex3f(wid,hei,0);
    glVertex3f(wid,-5,0);
  }
  if (!(whichpads & BOTTOMPAD)) {
    glVertex3f(0,0,0);
    glVertex3f(wid,0,0);
  }
  glEnd();
	  
  for(list<Stone *>::iterator s=stones.begin();s != stones.end();++s) {
    (*s)->render();
  }
  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    (*p)->render();
  }
  for (list<Ball *>::iterator b=balls.begin();b != balls.end();++b) {
    (*b)->render();
  }

  if (powerup) powerup->render();

  //  normCam(30); // the stones called lookAt from render()

  if (fadeout>0) {
    glColor4d(0,0,0,fadeout/FADETIME);
    glLoadIdentity();
    glBegin(GL_QUADS);
    glVertex3f(-1,-1,-1);
    glVertex3f(-1,+1,-1);
    glVertex3f(+1,+1,-1);
    glVertex3f(+1,-1,-1);
    glEnd();
  }
}

bool Level::update(double t) {

  /* Must first update the "heavy" objects (stuff on which e.g. the
     ball can bounce) and then "light" objects (stuff that
     bounces/hits stones) */

  // normCam(4);
  //int i;

  if (levelDone) {
    fadeout += t;
    if (fadeout>=FADETIME) {
      /* Pass to next level */
      fadeout=0;	    
      levelDone = false;
      //      delete level;
      levelnum++;

      resetCam();
      //      level = new Level();
      if (!load()) {
	printf("Can't open level %d (Did you just finish the game?? :)\n",levelnum);
	return false;
      }
      int mx,my;
      SDL_GetMouseState(&mx,&my);
      movePads(((double)mx)/screen_width,
	       ((double)my)/screen_height);
    }
  }

  /* Heavy objects */
  for(list<Stone *>::iterator s=stones.begin();s != stones.end();++s) {
    (*s)->update(t,this);
  }

  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    (*p)->update(t);
  }

  /* Light objects */
  int bcnt = 0;
  // int gone=-1;
  list<Ball *>::iterator b=balls.begin();
  while (b != balls.end()) {
    if ((*b)->update(t,this)) {
      ++bcnt;
      ++b;
    } else {
      // if ((gone==-1) && (bcnt==0)) {
      // 	gone=i; // deal with this one after the loop
      // } else {
      delete (*b);
      b = balls.erase(b);
      // }
    }
  }
  //  if (gone!=-1) {
  if (!bcnt) {
  // if (bcnt) {
  //   delete balls[gone];
  //   balls[gone]=0;
  // } else {
      /* Lost all balls. */
      /* Not counting lives for now (or ever?) */
    for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
      if (!(*p)->is_remote())
	(*p)->reset();
    }

  if (favpad && !passive)
    extra_ball(favpad);

      //      dontLook();
  }
  // }
  if (powerup) {
    if (powerup->update(t,this)) {
      delete powerup;
      powerup = 0;
    }
  }
  return true;
}

void Level::movePads(double x,double y) {
  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    /* We inverse the vertical parameter to have low value at the
       bottom */
    if (!((*p)->is_remote()))
      (*p)->move(wid*x,hei*(1-y));
  }
}

void Level::shiftPads(double dx,double dy) {
  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    /* We inverse the vertical parameter to have low value at the
       bottom */
    if ((*p)->shift(wid*dx,-hei*dy)) {
      for (list<Ball *>::iterator b=balls.begin();b != balls.end();++b) {
	(*b)->release(*p);
      }
    }
  }
}

void Level::releaseBalls() {
  for (list<Ball *>::iterator b=balls.begin();b != balls.end();++b) {
    (*b)->release();
  }
  for (list<Pad *>::iterator p=pads.begin();p != pads.end();++p) {
    (*p)->activate();
  }
  for(list<Stone *>::iterator s=stones.begin();s != stones.end();++s) {
    (*s)->activate();
  }
}

void Level::extra_ball(Pad *p) {
  // int slot=-1;
  // for (list<Ball *>::iterator b=balls.begin();b != balls.end();++b) {
  //   if (!(*b)) {
  //     slot = i;
  //     break;
  //   }
  // }
  // if (slot==-1) {
  //   Ball **na = new Ball*[ballcap+1];
  //   memcpy(na,balls,sizeof(Ball *) * ballcap);
  //   slot=ballcap;
  //   ballcap++;
  //   delete [] balls;
  //   balls=na;
  // }

  Ball *b = new Ball(0);
  add_ball(b);
  //  balls.push_back(b);

  b->init(0,0);
  b->attach(p);
}


// void Level::lost() {
//   /* Not counting lives for now (or ever?) */
//   for (int i=0;i<padcount;i++) {
//     (*p)->reset();
//   }
//   //  ball->reset();
//   favpad->capture(ball);
//   dontLook();
// }
