#include "moving.h"

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include <math.h>

#include "main.h"

Moving::Moving(int direction,bool unbreakable,unsigned int id) : Stone(id) {
  if (direction) {
    sx=0; sy=3;
  } else {
    sx=3; sy=0;
  }
  this->unbreakable = unbreakable;
}

Moving::~Moving() {}

void Moving::init(char ch,int x,int y) {
  Stone::init(ch,x,y);

  this->x = x;
  this->y = y;
  updating=false;
  dx=0; dy=0;
}

bool Moving::breakable() {
  return !unbreakable;
}

double Moving::getX() {
  return x;
}

double Moving::getY() {
  return y;
}

double Moving::check_collision(double x,double y,double dx,double dy, double *n, double *v,bool triggers) {
  if (updating) return 2; // a stone doesn't collide with itself ...

  /* If the stone is stuck somewhere, use the regular implementation */
  if (dx==0 && dy==0)
    return Stone::check_collision(x,y,dx,dy,n,v);

  /* Assume the stone is moving horizontally. (x,y) was the position
     when the stone was at (this->x-this->dx;this->y) but
     Stone::check_collision() assumes the stone doesn't move (i.e. has
     always been at (this->x;this->y). So we correct it by pretending
     the object was at (x+this->dx;y). The object's speed is similarly
     corrected. */
  double r = Stone::check_collision(x+this->dx,y+this->dy,
				    dx-this->dx,dy-this->dy,
				    n,v);
  if (r!=2) { // if there was a collision ...
    v[0] = sx; // ... record our speed
    v[1] = sy;
  }
  return r;
}

bool Moving::trigger(Littlecube *b) {
  if (unbreakable) {
    click[rand()%4].play();
    return false;
  } else
    return Stone::trigger(b);
}
void Moving::update(double t, Level *lvl) {
  Stone::update(t,lvl);

  double hx,hy; // what end of the stone will hit something first?

  updating = true;

  if (sx==0) {
    hx = 0.5;
    hy = sy>0? 0.99 : 0.01;
  } else { //if (sy==0) {
    /* general motion is not yet handled */
    hx = sx>0? 0.99 : 0.01;
    hy = 0.5;
  }

  if (lvl->check_collision(x+hx,y+hy,sx,sy,t,0,0,0,false) != 2) {
    /* the stone is about to hit something, stop, and move to the
       opposite direction next time */
    dx = 0;   dy = 0;
    sx = -sx; sy = -sy;
  } else {
    dx = t*sx; dy = t*sy;
    x += dx; y += dy;
  }

  updating = false;
}

void Moving::render() {
  if (unbreakable)
    glColor3f(0.7f,0.8f,0.9f);
  else
    glColor3f(0.5,1,0.5);

  render0();
}
