#ifndef PAD_H
#define PAD_H

/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

class Pad;
class Level;

#include "littlecube.h"
#include "cell.h"
#include "ball.h"

#define FIRECOUNT 10

/* Note, the first item has priority */
#define BOTTOMPAD 1
#define TOPPAD 2
#define LEFTPAD 4
#define RIGHTPAD 8
#define LASTPAD RIGHTPAD

#define CLS_PAD 3

struct pad_info {
  unsigned int id;
  unsigned char cls; // must be CLS_PAD
  unsigned char ev; /* EV_something */
  /* Those are the parameters passed to init(). With EV_MOVED, only x
     is used. */
  double x;
  double y;
  int direction;
};

class Pad : public Item {
 public:

  Pad(unsigned int id);
  ~Pad();

  /* the third argument must be one of the somethingPAD values. Note
     that those values indicate the normal position on the
     screen. The ball starts from the opposite position. */
  void init(Level *lvl,double x,double y,int direction);

  void process(struct item_info *data,size_t len);
  void poll();

  /* Return a pointer to a two int array containing [dirx,diry] (see
     private: below). Make a copy if you need that value for some
     time. */
  int *get_direction();

  /* Request the pad to move to that position. It won't go there
     instantly but, as update() gets called, gradually get closer to
     it */
  void move(double tx,double ty);

  /* Request the pad to move by the specified amount (call it every
     frame or the pad will stop moving).

     When requesting to shift a pad far towards the game area, any
     attached ball must be released. The return value is set to true
     when that happens. */
  bool shift(double dtx,double dty);

  /* Force the given ball to follow the pad is if it were "stuck" to
     it. Call this method continuously as long as the ball is stuck. */
  void capture(Ball *b);

  /* Notify the pad it got hit by a hard object (eg the ball) */
  void trigger();

  /* Tell the pad to activate whatever powers it has (eg fire)  */
  void activate();

  /* Make it wider */
  void grow();

  /* Make the pad able to fire stuff (use activate() to actually fire) */
  void fire();

  /* Remove all weird features obtained through powerups */
  void reset();

  /* See similarly named methods in level.h */
  double check_collision(double x,double y,double vx,double vy,double t, double *n, double *v);

  void render();

  void update(double t);

 private:

  Level *lvl;

  int direction; // somethingPAD
  /* dir{x,y}: Unit vector pointing from the pad to the game area */
  int dirx;
  int diry;

  double x;
  double targ; // target position for x (LEFT|RIGHT) or y (TOP|BOTTOM)
  double y;
  double angle;

  /* We do collision detection against both the current and previous
     positions to avoid having the ball getting through the pad */
  double prevx,prevy,prevangle;
  double dt; // how much time passed between the last two update()s

  double w; // half of width
  double tw; // "target w" so that pad growing is smooth

  bool can_fire; // whether the pad can fire stuff (use release() to actually fire)

  //  Ball *captured;

  Littlecube *firecube[FIRECOUNT];
};

#endif
