#ifndef REGION_H
#define REGION_H

/** A rectangular region of the level, used by the camera code to
    decide what to look at.

    Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */


class Region {
 public:
  Region();
  ~Region();

  /* Resets the region to a small bit in the bottom left corner */
  void reset();

  /* Enlarge the region to include the given point (if it wasn't
     included already */
  void include(double x,double y);

  /* Enlarge this region to contain the given one */
  void include(Region *that);

  /* Copy that region to this */
  void set_to(Region *that);

  void smooth_set_to(Region *that, double t);

  /* Return the coordinates of the region's centre as a pair of
     doubles (only valid until the next call to centre, make a copy!) */
  double *centre();

  double width();

  double height();

  double get_target_angle(bool vertical, double cam, double camz);

  double get_extraz(bool vertical, double cam, double camz, double ang);

  /* Draw a frame around the region (for debugging purposes) */
  void paint();

  /* Print the coordinates to stdout (for debugging purposes) */
  void print();

 private:
  bool neutral;

  /* First letter tells the dimension and second letter tells if it is
     the minimal ("a") or the maximal ("z") value */
  double xa,xz; 
  double ya,yz;

/* The Derivative of the previous values, if applicable */
  double dxa,dxz,dya,dyz;

};

#endif
