/** Copyright (C) 2010 Maxime Gamboni. See COPYING for
    copying/warranty info. */

#include <math.h>
#include <stdio.h>
#include <png.h>
#include <iostream>

#include "base.h" // for numerical constants

extern "C" { 
#include "jpeglib.h" 
}

using namespace std;

#include "tools.h"

double sq(double x) {
  return x*x;
}

/* Generic line intersection function, checks intersection with an
   horizontal obstacle but you can swap the meaning of x and y to
   test for a vertical obstacle. */
double wall_collision(double x,double y, /* trajectory start */
		     double dx,double dy, /* trajector delta */
		     double wx0, double wx1, /* obstacle x bounds */
		     double wy) { /* obstacle y coordinate */
  double d = (wy-y)/dy;
  if (d<0 || d>1) {
    return 2;
  } else {
    double wx = x+dx*d;
    if (wx<wx0 || wx>wx1)
      return 2;
    else {
      //      printf("[%f,%f + %f,%f | (%f-%f,%f) -> %f]",x,y,dx,dy,wx0,wx1,wy,d);
      return d;
    }
  }
}

void render_cube(double w,double h, double d, bool rounded, double rounding) {
    glBegin(GL_QUADS);

    //if (rounded) {glColor3f(0,0,0);} // debug

    double r=rounded?rounding:0;

    glNormal3f( 0, 0, 1); // front
    glVertex3f( r,   r,   d);
    glVertex3f( r,   h-r, d);
    glVertex3f( w-r, h-r, d);
    glVertex3f( w-r, r,   d);

    glNormal3f( 1, 0,   0); // right
    glVertex3f( w, r,   d-r);
    glVertex3f( w, h-r, d-r);
    glVertex3f( w, h-r, r);
    glVertex3f( w, r,   r);

    // glNormal3f( 0, 0,-1); // back (is this even ever shown?)
    // glVertex3f( w, 0, 0);
    // glVertex3f( w, h, 0);
    // glVertex3f( 0, h, 0);
    // glVertex3f( 0, 0, 0);

    glNormal3f(-1, 0, 0); // left
    glVertex3f( 0, h-r, r);
    glVertex3f( 0, r,   r);
    glVertex3f( 0, r,   d-r);
    glVertex3f( 0, h-r, d-r);

    glNormal3f( 0, 1, 0); // top
    glVertex3f( r  , h, r  );
    glVertex3f( w-r, h, r  );
    glVertex3f( w-r, h, d-r);
    glVertex3f( r  , h, d-r);

    glNormal3f( 0,-1, 0); // bottom
    glVertex3f( r  , 0, r  );
    glVertex3f( w-r, 0, r  );
    glVertex3f( w-r, 0, d-r);
    glVertex3f( r  , 0, d-r);

    if (rounded) {
      glNormal3f(0,1,0); // top back
      glVertex3f(w-r,h,  r);
      glVertex3f(r,  h,  r);
      glNormal3f(0,0,-1);
      glVertex3f(r,  h-r,0);
      glVertex3f(w-r,h-r,0);

      glNormal3f(0,0,1); // top front
      glVertex3f(r,  h-r,d );
      glVertex3f(w-r,h-r,d );
      glNormal3f(0,1,0);
      glVertex3f(w-r,h,  d-r);
      glVertex3f(r,  h,  d-r); 

      glNormal3f(0,-1,0); // bottom front
      glVertex3f(w-r,0,  d-r);
      glVertex3f(r,  0,  d-r); 
      glNormal3f(0,0,1);
      glVertex3f(r,  r,d );
      glVertex3f(w-r,r,d );

      glNormal3f(1,0,0); // right front
      glVertex3f(w,  r,  d-r);
      glVertex3f(w,  h-r,d-r);
      glNormal3f(0,0,1);
      glVertex3f(w-r,h-r,d);
      glVertex3f(w-r,r,  d);

      glNormal3f(0,0,1); // left front
      glVertex3f(r,r,  d);
      glVertex3f(r,h-r,d);
      glNormal3f(-1,0,0);
      glVertex3f(0,h-r,d-r);
      glVertex3f(0,r,  d-r);

      glNormal3f(1,0,0); // top right
      glVertex3f(w,  h-r,d-r);
      glVertex3f(w,  h-r,r  );
      glNormal3f(0,1,0);
      glVertex3f(w-r,h,  r  );
      glVertex3f(w-r,h,  d-r);

      glNormal3f(-1,0,0); // top left
      glVertex3f(0,  h-r,d-r);
      glVertex3f(0,  h-r,r  );
      glNormal3f(0,1,0);
      glVertex3f(r,h,  r  );
      glVertex3f(r,h,  d-r);

      glNormal3f(0,-1,0); // bottom left
      glVertex3f(r,0,r  );
      glVertex3f(r,0,d-r);
      glNormal3f(-1,0,0);
      glVertex3f(0,r,d-r);
      glVertex3f(0,r,r  );

      glNormal3f(0,-1,0); // bottom right
      glVertex3f(w-r,0,r  );
      glVertex3f(w-r,0,d-r);
      glNormal3f(1,0,0);
      glVertex3f(w  ,r,d-r);
      glVertex3f(w  ,r,r  );
    }

    glEnd();

    if (rounded) {
      glBegin(GL_TRIANGLES);
      glNormal3f(0,0,1); // top right
      glVertex3f(w-r,h-r,d);
      glNormal3f(1,0,0);
      glVertex3f(w,h-r,d-r);
      glNormal3f(0,1,0);
      glVertex3f(w-r,h,d-r);

      glNormal3f(0,0,1); // top left
      glVertex3f(r,h-r,d);
      glNormal3f(-1,0,0);
      glVertex3f(0,h-r,d-r);
      glNormal3f(0,1,0);
      glVertex3f(r,h,d-r);

      glNormal3f(0,0,1); // bottom right
      glVertex3f(w-r,r,d);
      glNormal3f(1,0,0);
      glVertex3f(w,r,d-r);
      glNormal3f(0,-1,0);
      glVertex3f(w-r,0,d-r);

      glNormal3f(0,0,1); // bottom left
      glVertex3f(r,r,d);
      glNormal3f(-1,0,0);
      glVertex3f(0,r,d-r);
      glNormal3f(0,-1,0);
      glVertex3f(r,0,d-r);
      glEnd();
    }
}

bool loadTexture(const char *file, GLuint *id) {
  // Returns in texture_id the first integer of MAX_NUM_TEXTURES consecutive
  // unused integers for texture.
  glGenTextures(1, id);

//   printf("created texture %d out of %s (error? %d)\n",*id,file,glGetError());

  struct  jpeg_decompress_struct cinfo;   // les infos du fichiers
  struct  jpeg_error_mgr jerr;            // les erreurs
  unsigned char *ligne;                   // une ligne (?)
  int  ImageSize;                         // Taille de l'image
  cinfo.err = jpeg_std_error( &jerr );
  jpeg_create_decompress( &cinfo );
  FILE *fichier = fopen( file,"rb" );  
  if ( fichier == NULL ) {
    cerr << "Error opening texture file." << endl;
    return false;
  }
  
  jpeg_stdio_src( &cinfo, fichier );
  jpeg_read_header( &cinfo, TRUE ); 
  jpeg_start_decompress ( &cinfo );  
  ImageSize = cinfo.image_width * cinfo.image_height * 3;//channels;
  
//   width = cinfo.image_width;
//   height = cinfo.image_height;
//     stride = cinfo.image_width*channels;    
  
  unsigned char *pData = new unsigned char [ImageSize]; /* are we supposed to free this after binding? */
  if ( pData == NULL ) {
    cerr << "Error allocating image memory" << endl;
    fclose (fichier);
    return false;    
  }
  
  ligne = pData;
  
  while ( cinfo.output_scanline < cinfo.output_height )
    {
      ligne = pData + cinfo.image_width * cinfo.output_scanline * 3; // channels
      jpeg_read_scanlines ( &cinfo, &ligne, 1 );
    }
  
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  
  fclose (fichier);
  
//   char pData[256*256*3];
//   int p=0;
//   for (int y=0;y<256;y++) for (int x=0;x<256;x++) {
//       pData[p++] = x&y;
//       pData[p++] = x&y;
//       pData[p++] = x&y;
//     }
  
  
  glBindTexture(GL_TEXTURE_2D, *id);

  glTexImage2D( GL_TEXTURE_2D, 0, 3, cinfo.image_width, cinfo.image_height, 
				0, GL_RGB, GL_UNSIGNED_BYTE, pData );

  delete pData;

  if (glGetError()) {
    printf("error in glTexImage2D: %d\n",glGetError());
    return false;
  }

  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT ); 
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

  return true;
}

bool loadTexture_png(const char *file, GLuint *id) {
  glGenTextures(1, id);

  int  ImageSize;                         // Taille de l'image

  FILE *fichier = fopen( file,"rb" );  
  if ( fichier == NULL ) {
    cerr << "Error opening texture file." << endl;
    return false;
  }

  /* I commented out the check that we're indeed dealing with a png
     file - I think libpng does this automatically when doing the
     actual loading anyway */

//   fread(header, 1, number, fp);
//   is_png = !png_sig_cmp(header, 0, number);
//   if (!is_png) {
//     printf("%s is not a png file\n",file);
//     return false;
//   }

/* (why do image loading libraries always have to be that complicated?
   What's wrong with just a data = load_my_picture(FILE *f); ?? ) */
  png_structp png_ptr = png_create_read_struct
    (PNG_LIBPNG_VER_STRING,0,0,0); // don't use custom error handling.

  if (!png_ptr) {
    fclose(fichier);
    return false;
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
      png_destroy_read_struct(&png_ptr,
			      (png_infopp)NULL, (png_infopp)NULL);
      fclose(fichier);
      return false;
  }

  png_infop end_info = png_create_info_struct(png_ptr);
  if (!end_info) {
      png_destroy_read_struct(&png_ptr, &info_ptr,
			      (png_infopp)NULL);
      fclose(fichier);
      return false;
  }
  png_init_io(png_ptr, fichier);
  png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

  fclose(fichier);

  png_byte **row_pointers = png_get_rows(png_ptr, info_ptr);
  int width = png_get_image_width(png_ptr,
				  info_ptr);
  int height=png_get_image_height(png_ptr,
				  info_ptr);
  ImageSize = width * height * 4;
  
//   printf("loaded png file %d x %d\n",width,height);

  unsigned char *pData = new unsigned char [ImageSize]; /* are we supposed to free this after binding? */
  if ( pData == NULL ) {
    cerr << "Error allocating image memory" << endl;
    png_destroy_read_struct(&png_ptr, &info_ptr,
			    (png_infopp)NULL);
    return false;    
  }
  
  /* I suspect the library doesn't allocate the rows in a single
     chunk, so we need to copy them in one. */
  for (int y=0;y<height;y++) {

  
//      for (int x=0;x<512;x++) {
//       pData[y*width+x*4+0] = x&y;
//       pData[y*width+x*4+1] = x&y;
//       pData[y*width+x*4+2] = x&y;
//       pData[y*width+x*4+3] = 128;
//     }

    memcpy(pData+width*4*y,row_pointers[y],width*4);
  }

  png_destroy_read_struct(&png_ptr, &info_ptr,
			      (png_infopp)NULL);

  glBindTexture(GL_TEXTURE_2D, *id);

  glTexImage2D( GL_TEXTURE_2D, 0, 4, width, height, 
				0, GL_RGBA, GL_UNSIGNED_BYTE, pData );
  if (glGetError()) {
    printf("error in glTexImage2D: %d\n",glGetError());
    return false;
  }

  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT ); 
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

  return true;
}

/**
 *	Replaces gluPerspective, sets the frustum to perspective mode.
 *	@param fovY	  : Field of vision in degrees in the y direction
 *	@param aspect : Aspect ratio of the viewport
 *	@param zNear  : The near clipping distance
 *	@param zFar	  : The far clipping distance
 */
void glPerspective( GLdouble fovY, GLdouble aspect, GLdouble zNear, GLdouble zFar )
{
	//	Very long (& in theory accurate!) version of Pi. Hopefully an optimizing compiler
	//	will replace references to this with the value!
	const GLdouble pi = 3.1415926535897932384626433832795;

	//	Half of the size of the x and y clipping planes.
	GLdouble fW, fH;

	//	Calculate the distance from 0 of the y clipping plane. Basically trig to calculate
	//	position of clipper at zNear.

	//	Note:	tan( double ) uses radians but OpenGL works in degrees so we convert
	//			degrees to radians by deviding by 180 then multiplying by pi.
	fH = tan( fovY / 180 * pi ) * zNear / 2;

	//	Calculate the distance from 0 of the x clipping plane based on the aspect ratio.
	fW = fH * aspect;

	//	Finally call glFrustum, this is all gluPerspective does anyway!
	//	This is why we calculate half the distance between the clipping planes - glFrustum
	//	takes an offset from zero for each clipping planes distance. (Saves 2 divides)
	glFrustum( -fW, fW, -fH, fH, zNear, zFar );
}
